/***
 * Jonathan: an Open Distributed Processing Environment 
 * Copyright (C) 2002 Kelua SA
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Contact: jonathan@objectweb.org
 *
 * Author: Christophe Hartwig
 *
 * -------------------------------------------------------------------
 * $ID$
 * -------------------------------------------------------------------
 * 
 */

package org.objectweb.jonathan.helpers;

import java.util.*;

import org.objectweb.jonathan.apis.kernel.Context;
import org.objectweb.jonathan.apis.kernel.Element;

/**
 * This class provides factory methods for the different kinds of Collections
 */
public class CollectionFactory {
   
   /**
    * Simple default constructor
    */
   public CollectionFactory() {}

   /**
    * Returns a new List using the given array of objects. Uses an ArrayList.
    * @param param the array to use to fill the list
    * @return the List
    */
   public List newList(Object [] param) {
      return newArrayList(param);
   }

   /**
    * Returns a new ArrayList using the given array of objects
    * @param param the array to use to fill the list
    * @return the ArrayList
    */
   public ArrayList newArrayList(Object [] param) {
      return toArrayList(param);
   }
   
   /**
    * Returns a new LinkedList using the given array of objects
    * @param param the array to use to fill the list
    * @return the LinkedList
    */
   public LinkedList newLinkedList(Object [] param) {
      return new LinkedList(toArrayList(param));
   }
   
   /**
    * Returns a new Vector using the given array of objects
    * @param param the array to use to fill the list
    * @return the Vector
    */
   public Vector newVector(Object [] param) {
      return new Vector(toArrayList(param));
   }
   
   /**
    * Returns a new Set using the given array of objects. Uses an HashSet.
    * @param param the array to use to fill the list
    * @return the Set
    */
   public Set newSet(Object [] param) {
      return newHashSet(param);
   }
   
   /**
    * Returns a new HashSet using the given array of objects
    * @param param the array to use to fill the list
    * @return the HashSet
    */
   public HashSet newHashSet(Object [] param) {
      return new HashSet(toArrayList(param));
   }
   
   /**
    * Returns a new TreeSet using the given array of objects
    * @param param the array to use to fill the list
    * @return the TreeSet
    */
   public TreeSet newTreeSet(Object [] param) {
      return new TreeSet(toArrayList(param));
   }

   /**
    * Returns a new TreeSet using the given array of objects
    * @param param the array to use to fill the list
    * @param aComparator the comparator used to sort elements in the tree
    * @return the TreeSet
    */
   public TreeSet newTreeSet(Object [] param, Comparator aComparator) {
      TreeSet aTreeSet = new TreeSet(aComparator);
      aTreeSet.addAll(toArrayList(param));
      return aTreeSet;
   }
   
   /**
    * Returns a new Hashtable
    * @param keys the vector making the keys
    * @param values the vector making the values
    * @return the Hashtable
    */
   public Hashtable newHashtable(Object [] keys, Object [] values) {
      return new Hashtable(newMap(keys, values));
   }
   
   /**
    * Returns a new Map. Uses a HashMap.
    * @param keys the vector making the keys
    * @param values the vector making the values
    * @return the Map
    */
   public Map newMap(Object [] keys, Object [] values) {
      return newHashMap(keys, values);
   }
   
   /**
    * Returns a new HashMap
    * @param keys the vector making the keys
    * @param values the vector making the values
    * @return the HashMap
    */
   public HashMap newHashMap(Object [] keys, Object [] values) {
      // sanity checks
      if ((keys == null) || (values == null)) {
         throw new IllegalArgumentException("null parameter to newHashMap");
      }
      if (keys.length != values.length) {
         throw new IllegalArgumentException("keys and values must have the same length to build a HashMap");
      }

      HashMap aHashMap = new HashMap();
      for (int count=0; count<keys.length; count++) {
         aHashMap.put(keys[count], values[count]);
      }

      return aHashMap;
   }
   
   /**
    * Returns a new TreeMap
    * @param keys the vector making the keys
    * @param values the vector making the values
    * @return the HashMap
    */
   public TreeMap newTreeMap(Object [] keys, Object [] values) {
      return new TreeMap(newHashMap(keys, values));
   }
   
   /**
    * Returns a new TreeMap
    * @param keys the vector making the keys
    * @param values the vector making the values
    * @param aComparator the comparator used to sort elements in the tree
    * @return the HashMap
    */
   public TreeMap newTreeMap(Object [] keys, Object [] values, Comparator aComparator) {
      TreeMap aTreeMap = new TreeMap(aComparator);
      aTreeMap.putAll(newHashMap(keys, values));
      return aTreeMap;
   }
   
   /**
    * Returns a new Set based on the contents of a Context. The elements
    * contained in the Context are added to the set
    * Uses an HashSet.
    * @param aContext the Context
    * @return the Set
    */
   public Set newSet(Context aContext) {
      return newHashSet(aContext);
   }
   
   /**
    * Returns a new HashSet based on the contents of a Context. The elements
    * contained in the Context are added to the set
    * @param aContext the Context
    * @return the HashSet
    */
   public HashSet newHashSet(Context aContext) {
      HashSet aSet = new HashSet();

      Enumeration all = aContext.getElements();
      while (all.hasMoreElements()) {
         Element each = (Element)all.nextElement();
         aSet.add(each.getValue());
      }
      return aSet;
   }

   /**
    * Returns a new HashMap based on the contents of a Context
    * @param aContext the Context
    * @return the HashMap
    */
   public HashMap newHashMap(Context aContext) {
      HashMap aHashMap = new HashMap();
      
      Enumeration all = aContext.getElements();
      while (all.hasMoreElements()) {
         Element each = (Element)all.nextElement();
         aHashMap.put(each.getName(), each.getValue());
      }
      return aHashMap;
   }
   
   /**
    * Returns a new Hashtable based on the contents of a Context
    * @param aContext the Context
    * @return the Hashtable
    */
   public Hashtable newHashtable(Context aContext) {
      return new Hashtable(newHashMap(aContext));
   }
   
   // utility to make a Collection out of a sequence
   private ArrayList toArrayList(Object [] param) {
      ArrayList aList = new ArrayList();

      for (int count=0;count<param.length;count++) {
         aList.add(param[count]);
      }
      return aList;
   }
   
}
