/***
 * Jonathan: an Open Distributed Processing Environment 
 * Copyright (C) 2000 France Telecom R&D
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Release: 3.0
 *
 * Contact: jonathan@objectweb.org
 *
 * Author: Bruno Dumant
 * 
 */

package org.objectweb.jonathan.libs.helpers;

/** 
 * Implements a hash table with integer or object keys and values.
 * Instance methods on HTable are NOT SYNCHRONIZED. It is up to the user 
 * to take the appropriate locks when necessary.
 */
public final class HTable {

   static private Object UNBOUND = new Object();

   /** table of holders */
   Holder[] table;
   /** size: number of registered objects in table */
   int size;
   /** table.length */
   int len;

   /** reuse holders */
   Holder unused;
   int unused_number;

   // context information
   Holder current;
   Holder previous;
   int index;

   /** 
    * Constructs a new HTable instance.
    */
   private HTable() {
      int initial_length = 101;
      table = new Holder[initial_length];
      len = initial_length;
      size = 0;
      unused_number = 0;
   }   

   /** 
    * Adds a new element to the target table.
    *<p>
    * This method is NOT synchronized: users need to manage the appropriate
    * monitors if necessary.
    * 
    * @param key    a (non-null) key to identify the element;
    * @param value  the (non-null) value of the element;
    * @return       the previous value, if any, otherwise null.
    */
   public Object put(Object key, Object value) {
      return put(key,Integer.MAX_VALUE,value);
   }

   /** 
    * Adds a new element to the target table.
    *<p>
    * This method is NOT synchronized: users need to manage the appropriate
    * monitors if necessary.
    *
    * @param key    a key to identify the element;
    * @param value  the (non-null) value of the element;
    * @return       the previous value, if any, otherwise null.
    */
   public Object put(int key, Object value) {
      return put(UNBOUND,key,value);
   }

   /** 
    * Adds a new element to the target table.
    *<p>
    * This method is NOT synchronized: users need to manage the appropriate
    * monitors if necessary.
    *
    * @param key    a (non-null) key to identify the element;
    * @param value  the value of the element;
    * @return       the previous value, if any, otherwise Integer.MAX_VALUE.
    */
   public int put(Object key,int value) {
      return put(key,Integer.MAX_VALUE,value);
   }

   /** 
    * Adds a new element to the target table.
    *<p>
    * This method is NOT synchronized: users need to manage the appropriate
    * monitors if necessary.
    *
    * @param key    a key to identify the element;
    * @param value  the value of the element;
    * @return       the previous value, if any, otherwise Integer.MAX_VALUE.
    */
   public int put(int key,int value) {
      return put(UNBOUND,key,value);
   }
   
   /** 
    * Removes the element identified by the provided key from the table.
    *<p>
    * This method is NOT synchronized: users need to manage the appropriate
    * monitors if necessary.
    *
    * @param key the element identifier.
    */
   public void remove(Object key) {
      setContext(key);
      removeCurrent();
   }

   /** 
    * Removes the element identified by the provided key from the table.
    *<p>
    * This method is NOT synchronized: users need to manage the appropriate
    * monitors if necessary.
    *
    * @param key the element identifier.
    */
   public void remove(int key) {
      setContext(key);
      removeCurrent();
   }
   
   /** 
    * Returns the element identified by key, or null if none exists, or if it has
    * been registered using {@link #put(Object,int) this put method}.
    *<p>
    * This method is NOT synchronized: users need to manage the appropriate
    * monitors if necessary.
    *
    * @param key an element identifier.
    * @return the corresponding element.
    */
   public Object get(Object key) {
      setContext(key);
      Object res = current != null ? current._value : null;
      current = null;
      previous = null;
      return res;
   }

   /** 
    * Returns the element identified by key, or null if none exists, or if it has
    * been registered using {@link #put(int,int) this put method}.
    *<p>
    * This method is NOT synchronized: users need to manage the appropriate
    * monitors if necessary.
    *
    * @param key an element identifier.
    * @return the corresponding element.
    */
   public Object get(int key) {
      setContext(key);
      Object res = current != null ? current._value : null;
      current = null;
      previous = null;
      return res;
   }
   
   /** 
    * Returns the element identified by key, or Integer.MAX_VALUE if none exists,
    * or if it has been registered using {@link #put(Object,Object) 
    * this put method}.    
    *<p>
    * This method is NOT synchronized: users need to manage the appropriate
    * monitors if necessary.
    *
    * @param key an element identifier.
    * @return the corresponding element.
    */
   public int getInt(Object key) {
      setContext(key);
      int res = current != null ? current._int_value : Integer.MAX_VALUE;
      current = null;
      previous = null;
      return res;
   }

   /** 
    * Returns the element identified by key, or Integer.MAX_VALUE if none exists,
    * or if it has been registered using {@link #put(int,Object) 
    * this put method}.    
    *<p>
    * This method is NOT synchronized: users need to manage the appropriate
    * monitors if necessary.
    *
    * @param key an element identifier.
    * @return the corresponding element.
    */
   public int getInt(int key) {
      setContext(key);
      int res = current != null ? current._int_value : Integer.MAX_VALUE;
      current = null;
      previous = null;
      return res;
   }

   /** 
    * Releases the target table. The table should no longer be used afterwards.
    *<p>
    * This method is NOT synchronized: users need to manage the appropriate
    * monitors if necessary.
    */
   public void release() {
      reset();
      release(this);
   }

   /** 
    * Removes all the elements from the target table.
    *<p>
    * This method is NOT synchronized: users need to manage the appropriate
    * monitors if necessary.
    */
   public void reset() {
      Holder holder;
      for (int i = 0; i < len; i++) {
         holder = table[i];
         while (holder != null) {
            table[i] = holder.next;
            holder.next = unused;
            unused = holder;
            holder._value = null;
            holder = table[i];
         }  
      }
      size = 0;
   }


   void setContext(Object key) {
      index = (key.hashCode() & 0x7FFFFFFF) % len;
      current = table[index];
      previous = null;
      while (! (current == null || current._key.equals(key))) {
         previous = current;
         current = current.next;
      }
   }

   void setContext(int key) {
      index = (key & 0x7FFFFFFF) % len;
      current = table[index];
      previous = null;
      while ((current != null) && (current._int_key != key)) {
         previous = current;
         current = current.next;
      }
   }

   Object put(Object key, int int_key, Object value) {
      if (key != UNBOUND) {
         setContext(key);
      } else {
         setContext(int_key);
      }
      if (current == null) {
         if (unused != null) {
            current = unused;
            unused = unused.next;
            unused_number--;
         } else {
            current = new Holder();
         }
         current._key = key;
         current._int_key = int_key;
         current._value = value;
         current._int_value = Integer.MAX_VALUE;
         current.next = table[index];
         table[index] = current;
         current = null;
         size++;
         if (size > len / 2) {
            rehash();
         }
         return null;
      } else {
         Object previous_value = current._value;
         current._value = value;
         current._int_value = Integer.MAX_VALUE;
         current = null;
         return previous_value;
      }
   }

 
   int put(Object key, int int_key, int int_value) {
      if (key != UNBOUND) {
         setContext(key);
      } else {
         setContext(int_key);
      }
      if (current == null) {
         if (unused != null) {
            current = unused;
            unused = unused.next;
            unused_number--;
         } else {
            current = new Holder();
         }
         current._key = key;
         current._int_key = int_key;
         current._value = null;
         current._int_value = int_value;
         current.next = table[index];
         table[index] = current;
         current = null;
         size++;
         if (size > len / 2) {
            rehash();
         }
         return Integer.MAX_VALUE;
      } else {
         int previous_value = current._int_value;
         current._int_value = int_value;
         current._value = null;
         current = null;
         return previous_value;
      }
   }


   void removeCurrent() {
      if (current != null) {
         size--;
         if (previous != null) {
            previous.next = current.next;
         } else {
            table[index] = current.next;
         }
         if (unused_number < len/2) {
            current.next = unused;
            unused = current;
            current._value = null;
         } else {
            current.next = null;
         }
      }  
   }

   void rehash() {
      //System.err.println("Rehashing " + this +  " " + len + " " + size);
      int new_len = 2 * len + 1;
      int index;
      Holder holder, next_holder;
      Holder[] new_table = new Holder[new_len];
      for (int i = 0; i < len; i++) {
         holder = table[i];
         while (holder != null) {
            next_holder = holder.next;
            // rehashing
            index = 
               ((holder._key == UNBOUND ? 
                 holder._int_key : holder._key.hashCode()) & 0x7FFFFFFF) % new_len;
            holder.next = new_table[index];
            new_table[index] = holder;
            //
            holder = next_holder; 
         }  
      }  
      table = new_table;
      len = new_len;
   } 

   static final class Holder {
      Object _key;
      int _int_key;
      Object _value;
      int _int_value;

      Holder next;
   }

   /**
    * POOL MANAGEMENT
    */

   static HTable[] pool = new HTable[10];
   static int pool_size = 0;   

   /** 
    * Returns a new HTable, taken from a pool of tables.
    * @return a new HTable, taken from a pool of tables.
    */
   static public synchronized HTable newTable() {
      if (pool_size > 0) {
         //System.err.println("Reusing " + pool[pool_size - 1]);
         return pool[--pool_size];
      } else {
         //System.err.println("new HTable ");
         return new HTable();
      }
   }
   
   static synchronized void release(HTable table) {
      //System.err.println("Releasing " + table);
      if (pool_size < pool.length) {
         //System.err.println("Storing " + table + " for reuse");
         pool[pool_size++] = table;
      }
   }
}


