/***
 * Jonathan: an Open Distributed Processing Environment 
 * Copyright (C) 2001 Kelua SA
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Contact: jonathan@objectweb.org
 *
 * Author: Bruno Dumant
 * 
 * -----------------------------------------------------------------
 * $Id: ClassPathFinderImpl.java,v 1.1.1.1 2002/08/06 13:13:49 bmi Exp $
 * -----------------------------------------------------------------
 */

package org.objectweb.jonathan.libs.kernel;

import org.objectweb.jonathan.apis.kernel.ClassPathFinder;

import org.objectweb.jonathan.libs.helpers.HTable;

import java.io.InputStream;
import java.io.IOException;
import java.net.URL;
import java.net.MalformedURLException;

import java.lang.reflect.Array;

public class ClassPathFinderImpl implements ClassPathFinder {


   public static ClassLoader system_class_loader = null;

   static {
      try {
         system_class_loader = (ClassLoader)
            ClassLoader.class.getDeclaredMethod("getSystemClassLoader",null).
            invoke(null,null);
      } catch (Exception ignored) {
         system_class_loader = null;
      }
   }

   ClassLoader class_loader;
   HTable cache;
   
   public ClassPathFinderImpl(ClassLoader _class_loader) {
      class_loader = _class_loader != null ? _class_loader : system_class_loader;
      cache = HTable.newTable();
   }
   
   public ClassLoader getClassLoader() {
      return class_loader;
   }
   
   public InputStream findResource(String _name) {

      InputStream resource_stream = null;
      
      resource_stream = class_loader != null ? class_loader.getResourceAsStream(_name) : null;
      // is it null? Try loading through a url
      if (resource_stream == null) {
         try {
            URL url = new URL(_name);
            resource_stream = url.openStream();
	 } catch (MalformedURLException ignored) {
	     // happens when the resource name doesn't specify a
	     // protocol, e.g. "jonathan.xml"
	     // this isn't a real exception...
         } catch (IOException e) {
             e.printStackTrace();
         }
      }

      // still null? Wouldn't it be on the local classloader instead?
      if (resource_stream == null) {
	  resource_stream = ClassPathFinderImpl.class.getClassLoader().getResourceAsStream(_name);
      }

      return resource_stream;
   }
      
   public Class findClass(String class_name) {
      Class result = (Class) cache.get(class_name);
      if (result != null) {
         return result;
      }

      int element_type_index = class_name.indexOf("[L");
      if (element_type_index != -1) {
         // class_name is an array type. Try to find element type first.
         String element_class_name = class_name.substring(element_type_index +
                                                          2,class_name.length()
                                                          -1);
         //System.err.println("Searching " + element_class_name);
         Class element_class = findClass(element_class_name);
         // compute the array size
         int size = 0;
         while (class_name.indexOf('[',size) != -1) {
            size++;
         }
         // create an array instance
         int[] sizes = new int[size];
         for (int i= 0; i < size; i++) {
            sizes[i] = 0;
         }
         Object array = Array.newInstance(element_class,sizes);
         // return its class
         return array.getClass();
      }

      if (class_loader != null){
         try{
            result = class_loader.loadClass(class_name);
         } catch (ClassNotFoundException ignored){
            //System.err.println("ClassPathFinderImpl: Exception loading class with " + class_loader + ": " + 
            //                   ignored);
         }
      }
      
      if (result == null){
         result = loadSystemClass(class_name,class_loader != system_class_loader);
      }

      if (result != null) {
         cache.put(class_name,result);
      }
      
      return result;
   }  
   
   public static Class loadSystemClass(String class_name) {
      return loadSystemClass(class_name,true);
   }  

   static Class loadSystemClass(String class_name,boolean cl_neq_scl) {

      Class result = null;
      
      if (system_class_loader != null && cl_neq_scl){
         try{
            result = system_class_loader.loadClass(class_name);
         } catch (ClassNotFoundException ignored){
            //System.err.println("ClassPathFinderImpl: Exception loading class with the system class loader: " +
            //                   ignored);
         }
      }
      
      if (result == null){
         try{
            result = Class.forName(class_name);
         } catch (ClassNotFoundException ignored){
            //System.err.println("ClassPathFinderImpl: Exception loading class with the Class.forName(): " +
            //                   ignored);
         }
      }

      if (result == null) {
         result = primitiveClass(class_name);
      }

      if (result == null) {
         System.err.println("WARNING: Could not load class " + class_name);
      }
      return result;
   }  

   static Class primitiveClass(String name) {
      switch (name.charAt(0)) {
         case 'b':
            switch (name.charAt(1)) {
               case 'o':
                  return "boolean".equals(name) ? boolean.class : null;
               default:
                  return "byte".equals(name) ? byte.class : null;
            }
         case 'c':
            return "char".equals(name) ? char.class : null;
         case 'd':
            return "double".equals(name) ? double.class : null;
         case 'f':
            return "float".equals(name) ? float.class : null;
         case 'i':
            return "int".equals(name) ? int.class : null;
         case 'l':
            return "long".equals(name) ? long.class : null;
         case 's':
            return "short".equals(name) ? short.class : null;
         default:
            return "void".equals(name) ? void.class : null;
      }
   }   


}



