/***
 * Jonathan: an Open Distributed Processing Environment 
 * Copyright (C) 2000 France Telecom R&D
 * Copyright (C) 2001 Kelua SA
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Contact: jonathan@objectweb.org
 *
 * Author: Fabien Delpiano, Bruno Dumant
 * 
 * --------------------------------------------------------------------------
 * $Id: JContextFactory.java,v 1.1.1.1 2002/08/06 13:13:50 bmi Exp $
 * -------------------------------------------------------------------------- 
 */

package org.objectweb.jonathan.libs.kernel;

import org.objectweb.jonathan.apis.kernel.Context;
import org.objectweb.jonathan.apis.kernel.ContextFactory;
import org.objectweb.jonathan.apis.kernel.Element;
import org.objectweb.jonathan.apis.kernel.Name;
import org.objectweb.jonathan.apis.kernel.Component;
import org.objectweb.jonathan.apis.kernel.JonathanException;
import org.objectweb.jonathan.apis.kernel.InternalException;

import java.util.Enumeration;
import java.util.NoSuchElementException;

import org.objectweb.jonathan.libs.helpers.NameHelpers;

/** 
 * Implementation of {@link ContextFactory ContextFactory}.
 */
public class JContextFactory implements ContextFactory {


   static JContext[] pool = new JContext[10];
   static int pool_size = 0;   

   // number of created JContexts (just for identification).
   static int nb;

   /** 
    * Indicates whether type checking should be performed.
    */
   public static final boolean check_types = false;

   public static final JContextFactory instance = new JContextFactory();

   /** 
    * Returns a new {@link JContextFactory.JContext JContext} 
    * instance.
    * @return a new {@link JContextFactory.JContext} instance.
    */
   public Context newContext() {
      synchronized (JContextFactory.class) {
         if (pool_size > 0) {
            JContext context = pool[--pool_size];
            context.acquire();
            return context;
         } else {
            return new JContext();
         }
      }
   }

   /** 
    * Returns a new {@link JContextFactory.JContext JContext} 
    * instance with scope 
    * <code>scope</code>. 
    * <p>
    * <code>_scope</code> must have been acquired by the caller first, if it is
    * to be shared.
    *
    * @param scope the scope of the returned context instance;
    * @return a new {@link JContextFactory.JContext JContext} 
    *         instance.
    */
   public Context newContext(Context _scope) {
      synchronized (JContextFactory.class) {
         if (pool_size > 0) {
            JContext context =  pool[--pool_size];
            context.acquire();
            context.enclosing_context = _scope;
            return context;
         } else {
            return new JContext(_scope);
         }
      }
   }

   /** 
    * Implementation of {@link Context Context}.
    */
   protected class JContext extends JComponent implements Context {

      /** table of holders */
      ContextElement[] table;
      /** size */
      int size;
      
      int ref_count;
      
      // pool of context elements
      ContextElement unused = null;
      
      /** The enclosing context (scope) */
      protected Context enclosing_context = null;
      
      /** Identifies the target context */
      protected final int id;

      /** 
       * Constructs a new JContext instance.
       */
      protected JContext() {
         table = new ContextElement[7];
         size = 0;
         id = nb++;
         ref_count = 1;
         type = this.getClass();
         value = this;
      }
      
      /** 
       * Creates a new JContext with the specified enclosing context.
       * @param _scope a Context.
       */
      protected JContext(Context _scope) {
         this();
         enclosing_context = _scope;
      }
      
      /** 
       * may need to be synchronized
       * @param _name
       * @return the element corresponding to <code>_name</code>, if any.
       */
      final public Element getElement(Name _name) {
         //System.err.println("Looking for: " + _name);
         Element $element = getElement(_name.getId(),false);
         //System.err.println("Found \"" + _name.getId() + "\": " + $element);
         Name $subname = _name.getSubName();
         if ($element == null || $subname == null) {
            return $element;
         } else {
            Component $component = $element.getComponent();
            //System.err.println("Looking for: " + $subname);
            return $component != null ? $component.getElement($subname) : null;
         }
      }

      final public int getIntValue(Name _name) {
         Element element = getElement(_name);
         if (element != null) {
            return element.getIntValue();
         } else {
            return Integer.MAX_VALUE;
         }
      }
      
      final public Element getElement(String _name,char _separator) {
         if (_separator == 0) {
            return getElement(_name,false);
         } else {
            return getElement(NameHelpers.newName(_name,_separator));
         }
      }
      
      final public Component getComponent(String name,char separator) {
         if (separator == 0) {
            Element element = getElement(name,false);
            if (element != null) {
               return element.getComponent();
            } else {
               return null;
            }
         } else {
            return getComponent(NameHelpers.newName(name,separator));
         }
      }

      final public Object getFactoryValue(String name,char separator) throws JonathanException {
         Component $comp = getComponent(name,separator);
         return $comp.getFactoryValue();
      }
      

      final public Object getValue(String name,char separator) {
         if (separator == 0) {
            Element element = getElement(name,false);
            if (element != null) {
               return element.getValue();
            } else {
               return NO_VALUE;
            }
         } else {
            return getValue(NameHelpers.newName(name,separator));
         }
      }
      
      final public int getIntValue(String name,char separator) {
         if (separator == 0) {
            Element element = getElement(name,false);
            if (element != null) {
               return element.getIntValue();
            } else {
               return Integer.MAX_VALUE;
            }
         } else {
            return getIntValue(NameHelpers.newName(name,separator));
         }
      }      
         
      public final Element addElement(Name _name, Component _component) 
         throws JonathanException {
         JContext $local_context = (JContext) addOrGetContext(NameHelpers.getPrefix(_name));
         return $local_context.addLocalElement(NameHelpers.getSuffix(_name).getId(),
                                               _component,check_types);
      }

      public final Element addElement(Name _name, Class _type, Object _value) 
         throws JonathanException {
         return addElement(_name,newComponent(_type,_value));
      }
      
      public final Element addElement(Name _name, Class _type, int _value) 
         throws JonathanException {
         return addElement(_name,newComponent(_type,_value));
      }

      public final Element addElement(String _name,Component _component,char separator) 
         throws JonathanException {
         if (separator != 0) {
            return addElement(NameHelpers.newName(_name,separator),_component);
         } else {
            return addLocalElement(_name,_component,check_types);
         }
      }

      public final Element addElement(String name,Class type,Object value,char separator) 
         throws JonathanException {
         if (separator != 0) {
            return addElement(NameHelpers.newName(name,separator),newComponent(type,value));
         } else {
            return addLocalElement(name,newComponent(type,value),check_types);
         }
      }

      public final Element addElement(String name, Class type, int value,char separator) 
         throws JonathanException {
         if (separator != 0) {
            return addElement(NameHelpers.newName(name,separator),newComponent(type,value));
         } else {
            return addLocalElement(name,newComponent(type,value),false);
         }
      }

      public Context addOrGetContext(String _name,char _separator) throws JonathanException {
         return addOrGetContext(NameHelpers.newName(_name,_separator));
      }
         
      public Context addOrGetContext(Name _name) throws JonathanException {
         if (_name == null) {
            return this;
         }
         String $id = _name.getId();
         Name $subname = _name.getSubName();
         JContext $local_context = addOrGetLocalContext($id);
         return $local_context.addOrGetContext($subname);
      }      

      public Enumeration getElements() {
         return new InnerEnumeration();
      }

      public Context getScope() {
         return enclosing_context;
      }

      public Component internalFork() {
         Component result = JContextFactory.this.newContext(this);
         result.setReference(reference);
         return result;
      }
      

      public Context setScope(Context scope) {
         Context previous_scope = enclosing_context;
         enclosing_context = scope;
         return previous_scope;
      }

      /** 
       * needs to be synchronized.
       */
      public void reset() {
         int len = table.length;
         ContextElement current = null;
         ContextElement next = null;
         for (int i = 0; i  < len; i++) {
            current = table[i];
            while (current != null) {
               next = current.next;
               current.reset();
               current.next = unused;
               unused = current;
               current = next;
            }
            table[i] = null;
         }
         size = 0;
      }
      
      public void acquire() {
         ref_count++;
      }

      public void release() {
         ref_count--;
         if (ref_count == 0) {
            if (enclosing_context != null) {
               enclosing_context.release();
            }
            enclosing_context = null;
            reset();
            reference = null;
            recycle();
         }
      }
      
      public Component duplicate() {
         acquire();
         return this;
      }


      public String toString() {
         return "JContext[" + id + (enclosing_context != null ? "," + enclosing_context + "]]" : "]");
      }

      protected void recycle() {
         synchronized (JContextFactory.class) {
            if (pool_size < pool.length) {
               pool[pool_size++] = this;
            }
         }
      }

      /** 
       * Returns the local element identified by name, or null if none exists.
       *
       * @param name a local name;
       * @param local if true, don't search the enclosing context.
       * @return the local element identified by name, or null if none exists.
       */
      protected Element getElement(String name,boolean local) {   
         int index = (name.hashCode() & 0xFFFFFFF) % table.length;
         ContextElement holder = table[index];
         while (! (holder == null || holder.name.equals(name))) {
            holder = holder.next;
         }
         if (holder != null) {
            return holder;
         } else {
            if ((! local) && enclosing_context != null){
               return enclosing_context.getElement(name,(char) 0);
            } else {
               return null;
            }  
         }
      }

      /** 
       * Adds an Element to the Context if no other exists with the same name.
       * <p>
       * must be  called in a synchronized context
       *
       * @param name the name of the new Element.
       * @param type the declared type of the new Element.
       * @param value the value of the Element to be inserted.
       * @return the Element already present in the Context with the same name, 
       *         null if none existed.
       * @exception JonathanException
       */
      protected Element addLocalElement(ContextElement element,
                                        boolean _check_types)
         throws JonathanException {
         String name = element.getName();
         Class type = null;
         
         if (check_types) {
            type = element.getType();
            if (! element.checkType()) {
               throw new JonathanException("Type error: value " + 
                                           element.getValue() + 
                                           " is not of type " + type);
            }
            
            if (enclosing_context != null) {
               Element $scope_element = enclosing_context.getElement(name,(char) 0);
               if ($scope_element != null && type != null) {
                  Class $scope_type = $scope_element.getType();
                  if (type != $scope_type && ! $scope_type.isAssignableFrom(type)) {
                     throw new JonathanException(type + " is not a subtype of " + 
                                                 $scope_type);
                  }
               }
            }
         }
         
         int len = table.length;
         int index = (name.hashCode() & 0xFFFFFFF) % len;
         ContextElement current = table[index];
         ContextElement previous = null;

         while (! (current == null || current.name.equals(name))) {
            previous = current;
            current = current.next;
         }   
         
         if (check_types && current != null) {
            Class $current_type = current.getType();
            if (type != $current_type && ! $current_type.isAssignableFrom(type)) {
               throw new JonathanException(type + " is not a subtype of " + 
                                           $current_type);
            }
         }

         ContextElement $toBeReturned = null;
         
         if (current == null) {
            element.next = table[index];
            table[index] = element;
            size++;
            if (size > len / 2) {
               rehash(len);
            }
         } else {
            $toBeReturned = current;
            element.next = current.next;
            current.next = null;
            if (previous != null) {
               previous.next = element;
            } else {
               table[index] = element;
            }
         }
         return $toBeReturned;
      }

      /** 
       * Creates a new empty context element.
       * @return a new context element.
       */
      protected ContextElement newContextElement(Context _container) {
         ContextElement result = null;
         if (unused != null) {
            result = unused;
            unused = unused.next;
            result.next = null;
            result.container = _container;
         } else {
            result = new ContextElement(_container);
         }
         return result;
      }

      /** 
       * Creates a new context element, using the provided parameters
       * to initialize it.
       * @param name a name;
       * @param type a type; 
       * @param value an object reference value; 
       * @param int_value an integer value; 
       * @return a new context element, initialized using the provided parameters
       */
      protected final ContextElement newContextElement(String name,Component component,Context container) {
         ContextElement result = newContextElement(container);
         result.name = name;
         result.component = component;
         component.setReference(result);
         return result;
      }

      protected Component newComponent(Class type, Object value) {
         return new JComponent(type,value);
      }
      
      protected Component newComponent(Class type, int value) {
         return new JComponent(type,value);
      }
      

      protected void finalize() throws Throwable {
         reset();
      }
      
      protected JContext addOrGetLocalContext(String _id) 
         throws JonathanException {

         Element $previous_element = getElement(_id,true);
         Object $previous_value = null;
         JContext $local_context = null;
         
         if ($previous_element == null) {
            if (enclosing_context != null) {
               Element $scope_element = enclosing_context.getElement(_id,(char) 0);
               if ($scope_element != null) {
                  $previous_value = $scope_element.getValue();
                  if ($previous_value == NO_VALUE ||
                      $previous_value == null) {
                     $local_context = (JContext) newContext();
                  } else {
                     try {
                        /// The enclosing element must be a context: we create a 
                        /// local context in its scope, and we add the sub element
                        /// to it.
                        Context $scope_context = (Context) $previous_value;
                        $scope_context.acquire();
                        $local_context = (JContext) newContext($scope_context);
                     } catch (ClassCastException e) {
                        throw new JonathanException("Non context element " + 
                                                    $previous_value + 
                                                    "found on path.");
                     }
                  }
               } else {
                  $local_context = (JContext) newContext();
               }
            } else {
               $local_context = (JContext) newContext();
            }                     
            addLocalElement(_id,$local_context, false);
         } else if (($previous_value = $previous_element.getValue()) 
                    instanceof Context) {
            $local_context = (JContext) $previous_value;
         } else {
            $local_context = (JContext) newContext();
            addLocalElement(_id,$local_context,false);
         }
         return $local_context;
      }
      
      final Element addLocalElement(String name,Component component,boolean _check_types)
         throws JonathanException {
         return addLocalElement(newContextElement(name,component,this),
                                _check_types);
      }

      final void rehash(int len) {
         int new_len = 2 * len + 1;
         int index;
         ContextElement holder, next_holder;
         ContextElement[] new_table = new ContextElement[new_len];
         for (int i = 0; i < len; i++) {
            holder = table[i];
            while (holder != null) {
               next_holder = holder.next;
               // rehashing
               index = (holder.name.hashCode() & 0xFFFFFFF) % new_len;
               holder.next = new_table[index];
               new_table[index] = holder;
               //
               holder = next_holder;
            }
         }
         table = new_table;
      }
      
      protected Element handleScopeElement(Element element) {
         return element;
      }

      protected class InnerEnumeration implements Enumeration {
         // current_index is the first unconsidered index in the table
         int current_index = 0;
         Element next_element = null;
         Enumeration scope_elements = null;
         
         protected InnerEnumeration(){
            next_element = getNextElement();
         }
         
         public boolean hasMoreElements(){
            return next_element != null;
         }
         
         public Object nextElement() {
            if (next_element == null) {
               throw new NoSuchElementException();
            }
            Element next = next_element;
            next_element = getNextElement();
            return next;
         }
         
         protected Element getNextElement() {
            ContextElement current = (ContextElement) next_element;
            if (current != null && current.next != null && scope_elements == null){
               return current.next;
            }
            
            if (scope_elements == null && current_index < table.length) {   
               current = null;
               while (current == null && current_index < table.length) {
                  current = table[current_index];
                  current_index++;
               }
               
               if (current != null) {
                  return current;
               }
            }  
         
            Element scope_current = null;
            
            if (scope_elements == null) {
               if (enclosing_context != null){
                  scope_elements = enclosing_context.getElements();
               }
            }
            
            
            if (scope_elements == null){
               return null;
            } else {
               while (scope_elements.hasMoreElements()) {
                  scope_current = (Element) scope_elements.nextElement();
                  if (getElement(scope_current.getName(),true) == null) {
                     return handleScopeElement(scope_current);
                  }
               }
               return null;
            }
         }         
      }   
   }
   
   /** 
    * Class of the context elements used by JContextFactory.
    */
   static public class ContextElement extends JElement {   

      /** next in the linked list */
      ContextElement next;
      
      /** 
       * Constructs a new context element.
       */
      public ContextElement(Context container) {
         this.container = container;
      }
      
      /** 
       * Contructs a new context element, using the provided component.
       * After contruction, the containing element of the provided component is 
       * the newly created element.
       * @param _component a component.
       */
//        public ContextElement(Component _component) {
//           component = _component;
//           if (_component != null) {
//              component.setReference(this);
//           }
//        }
      
      public String toString() {
         return "JContextElement[" + name + " " + component + "]";
      }

      protected boolean checkType() {
         Object $value = getValue();
         Class $type = getType();
         return 
            $value == null || $value == Context.NO_VALUE  || $type.isInstance($value);
      }
      
      /** 
       * Resets the element, setting all object reference members to null, 
       * and possibly recycling the element in a pool.
       */
      protected void reset() {
         next = null;
         name = null;
         component = null;
         container = null;
      }
   }
}
