/***
 * Jonathan: an Open Distributed Processing Environment 
 * Copyright (C) 2001 Kelua SA
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Contact: jonathan@objectweb.org
 *
 * Author: Priscille Durville
 *
 * -------------------------------------------------------------------
 * $Id: DecoderHelper.java,v 1.2 2002/09/26 08:36:21 schambon Exp $
 * -------------------------------------------------------------------
 * 
 */

package org.objectweb.jonathan.libs.kernel.decoders;

import org.objectweb.jonathan.apis.kernel.ClassPathFinder;
import org.objectweb.jonathan.apis.kernel.Component;
import org.objectweb.jonathan.apis.kernel.Context;
import org.objectweb.jonathan.apis.kernel.JonathanException;
import org.objectweb.jonathan.apis.kernel.Name;
import org.objectweb.jonathan.apis.kernel.Instanciable;

import org.objectweb.jonathan.libs.helpers.NameHelpers;
import org.objectweb.jonathan.libs.helpers.TraceHelpers;
import org.objectweb.jonathan.libs.kernel.ClassPathFinderImpl;

import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;

import java.util.List;
import java.util.ListIterator;

/** 
 * Helper class providing methods used by JImpicitFactory, KAtom and KAssemblage.
 */
public class DecoderHelper {

   static Name CLASSPATH_FINDER_NAME = NameHelpers.newName("/classpath_finder",'/');

  /**
    * Returns the <code>_classe</code> java class's Method whose name is _name
    * and arguments'types are listed in <code>_arguments</code> as JArguments. 
    * To reach this goal, this method first tries to find and load each of
    * <code>_arguments</code>' java classes
    * with the ClasspathFinder found at <B>"/classpath_finder"</B> (in the
    * <code>_context</code> context).
    * @param _class the belonging class of the method
    * @param _name the name of the method
    * @param _arguments the parameters of the method
    * @param _context the context in which find the ClasspathFinder
    * @return its Class method
    * @exception JonathanException if something goes wrong.
    */
   public static Method getMethod(Class _class,String _name,List _arguments,Component _context) 
     throws JonathanException {

      //1.attempt to load all the necessary classes:
      Object $component = _context.getValue(CLASSPATH_FINDER_NAME);
      ClassPathFinder $class_finder = 
         $component instanceof ClassPathFinder ? (ClassPathFinder) $component : null;
         
      ListIterator $iterator = _arguments.listIterator();
      int $nb_parameters = _arguments.size();
      Class[] $parameter_types = new Class[$nb_parameters];
      String $parameter_type = null;
      for (int i = 0; i < $nb_parameters; i++) {
         $parameter_type = ((JArgument) $iterator.next()).type;
         if (isArrayClassName($parameter_type)) {
            $parameter_type = getClassNameForArray($parameter_type);
         }
         if ($class_finder != null) {
            $parameter_types[i] = $class_finder.findClass($parameter_type);
         } else {
            $parameter_types[i] = ClassPathFinderImpl.loadSystemClass($parameter_type);
         }
         if ($parameter_types[i] == null) {
            throw new JonathanException("ERROR in implicit factory: could not call method" + _name);
         }
      }

      //2.attempt to create the method
      try {
         Method $method = _class.getMethod(_name,$parameter_types);
         return $method;
      } catch (SecurityException e) {
         throw new JonathanException(e);
      } catch (NoSuchMethodException e) {
         throw new JonathanException(e);
      }
   }

   /**
    * Returns the <code>_classe</code> java class's Constructor whose arguments'types
    * are listed in <code>_arguments</code> as JArguments. To reach this goal, this
    * method first tries to find and load <code>_classe</code> and each of
    * <code>arguments</code>' java classes with the ClasspathFinder found at
    * <B>"/classpath_finder"</B> (in the
    * <code>_context</code> context).
    * @param _classe the class's...
    * @param _arguments ...signature
    * @return its Class constructor
    * @exception JonathanException if something goes wrong.
    */
   public static Constructor getConstructor(String _classe,List _arguments,Component _context)
      throws JonathanException {  

      Object $component = _context.getValue(CLASSPATH_FINDER_NAME);
      ClassPathFinder $class_finder = 
         $component instanceof ClassPathFinder ? (ClassPathFinder) $component : null;         
      Class $implementation_class = null;

      if ($class_finder != null) {
         $implementation_class = $class_finder.findClass(_classe);
      } else {
         $implementation_class = ClassPathFinderImpl.loadSystemClass(_classe);
      }
         
      if ($implementation_class == null) {
         throw new JonathanException("ERROR in implicit factory: could not instantiate class " + _classe );
      }
         
      ListIterator $iterator = _arguments.listIterator();
      int $nb_parameters = _arguments.size();
      Class[] $parameter_types = new Class[$nb_parameters];
      String $parameter_type = null;
      for (int i = 0; i < $nb_parameters; i++) {
         $parameter_type = ((JArgument) $iterator.next()).type;
         if (isArrayClassName($parameter_type)) {
            $parameter_type = getClassNameForArray($parameter_type);
         }
         if ($class_finder != null) {
            $parameter_types[i] = $class_finder.findClass($parameter_type);
         } else {
            $parameter_types[i] = ClassPathFinderImpl.loadSystemClass($parameter_type);
         }
         if ($parameter_types[i] == null) {
            throw new JonathanException("ERROR in implicit factory: could not instantiate type " + _classe );
         }
      }
      
      try {
         Constructor $constructor = $implementation_class.getConstructor($parameter_types);
         return $constructor;
      } catch (SecurityException e) {
         throw new JonathanException(e);
      } catch (NoSuchMethodException e) {
         throw new JonathanException(e);
      }
   }

   /**
    * "applies" the <B>_context</B> Context's elements on the
    * specified method whose signature is
    * <B>_method(_arguments)</B>, i.e.
    * <UL> <LI>find and instantiates <B>_arguments</B>' JArguments</LI>
    * <LI>calls <B>_method</B> with them as its
    * parameters on the object _object</LI></UL>
    * @param _object the object calling method
    * @param _method method to call
    * @param _arguments its arguments
    * @param _context contains the instanciation parameters
    * @exception JonathanException if something goes wrong.
    */
   public static Object apply(Object _object, Method _method, List _arguments, Context _context,Instanciable _instance) 
      throws JonathanException {
         
      int $nb_parameters = _arguments.size();
      Object[] $arguments = new Object[$nb_parameters];
      Class[] $argument_types = _method.getParameterTypes();
      ListIterator $iterator = _arguments.listIterator();
      String $name = null;
      Object $arg = null;
      for (int i = 0; i < $nb_parameters; i++) {
         $name = ((JArgument) $iterator.next()).name;
         $arg = _context.getValue($name,'/');
         if ($arg == Context.NO_VALUE) {
            throw new JonathanException("ERROR in factory: missing argument : " + $name);
         } else if ($argument_types[i].isInstance($arg)) {
            $arguments[i] = $arg;
         } else if ($argument_types[i].isPrimitive()) {
            $arguments[i] = $arg; // no check yet
         } else if ($arg == null) {
            $arguments[i] = $arg;
         } else {
            throw new JonathanException("ERROR in factory: argument " + $arg + " named " + $name +
                                        " is not of type " + $argument_types[i]);
         }
      }

      try {
         if ((_instance == null) || (! _instance.isInstanciated())){
            return _method.invoke(_object, $arguments);
         } else {
            return _instance.getInstance();
         }
      } catch (InvocationTargetException e) {
         Throwable $target_exception = e.getTargetException();
         if ($target_exception instanceof JonathanException) {
            throw (JonathanException) $target_exception;
         }
         if ($target_exception instanceof Exception) {
            throw new JonathanException((Exception) $target_exception);
         }
         throw new JonathanException("ERROR in factory invocation : " +
                                     TraceHelpers.convertStackToString($target_exception));

      } catch (Exception e) {
         throw new JonathanException(e);
      }
   }

   /**
    * "applies" the <B>_context</B> Context's elements on the
    * implementation alternative whose signature is
    * <B>_constructor(_arguments)</B>, i.e.
    * <UL> <LI>find and instantiates <B>_arguments</B>' JArguments</LI>
    * <LI>instantiates <B>_constructor</B> with them as its
    * parameters</LI></UL>
    * @param _constructor constructor to call
    * @param _arguments its arguments
    * @param _context contains the instanciation parameterd
    * @return the created instance
    * @exception JonathanException if something goes wrong.
    */
   public static Object apply(Constructor _constructor, List _arguments, Context _context,Instanciable _instance) 
      throws JonathanException {
         
      int $nb_parameters = _arguments.size();
      Object[] $arguments = new Object[$nb_parameters];
      Class[] $argument_types = _constructor.getParameterTypes();
      ListIterator $iterator = _arguments.listIterator();
      String $name = null;
      Object $arg = null;
      for (int i = 0; i < $nb_parameters; i++) {
         $name = ((JArgument) $iterator.next()).name;
         $arg = _context.getValue($name,'/');
         if ($arg == Context.NO_VALUE) {
            throw new JonathanException("ERROR in factory: missing argument : " + $name);
         } else if ($argument_types[i].isInstance($arg)) {
            $arguments[i] = $arg;
         } else if ($argument_types[i].isPrimitive()) {
            $arguments[i] = $arg; // no check yet
         } else if ($arg == null) {
            $arguments[i] = $arg;
         } else {
            throw new JonathanException("ERROR in factory: argument " + $arg + " named " + $name +
                                        " is not of type " + $argument_types[i]);
         }
      }

      try {
         if ((_instance == null) || (! _instance.isInstanciated())){
            return _constructor.newInstance($arguments);
         } else {
            return _instance.getInstance();
         }

      } catch (InvocationTargetException e) {
         Throwable $target_exception = e.getTargetException();
         if ($target_exception instanceof JonathanException) {
            throw (JonathanException) $target_exception;
         }
         if ($target_exception instanceof Exception) {
            throw new JonathanException((Exception) $target_exception);
         }
         throw new JonathanException("ERROR in factory instanciation : " +
                                     TraceHelpers.convertStackToString($target_exception));

      } catch (Exception e) {
         throw new JonathanException(e);
      }
   }

   /** 
    * Test if the <code>_type</code> given is an array class name.
    * @param _type the type to be tested
    * @return true if it is an array class name, false otherwise.
    */
   public static boolean isArrayClassName(String _type) {
      return _type.endsWith("[]");
   }
   
   /** 
    * Return the class name corresponding to the <code>_array</code> given.
    * @param _array the array to be studied
    * @return this array class name.
    */
   public static String getClassNameForArray(String _array) {
      int dimension = -1;
      int indexOfBracket = -1;
      while ((indexOfBracket = _array.indexOf("[]", indexOfBracket +1 )) != -1) {
	 dimension++;
      }


      // remove the trailing []
      String $type_name = _array.substring(0, _array.indexOf("[]"));
      //String $type_name = _array.substring(0,_array.length()-2);
   
      StringBuffer sbuf = new StringBuffer();
      for (int i = 0; i < dimension; i++) sbuf.append("[");

      // try prim arrays first
      String $class_name = (String)JSequence.ARRAY_TYPES.get($type_name);

      if ($class_name == null) {
         // not a prim array
         $class_name = sbuf.toString() + "[L".concat($type_name).concat(";");
      } else {
	 $class_name = sbuf.toString() + $class_name;
      }

      return $class_name;
   }
}
