/***
 * Jonathan: an Open Distributed Processing Environment 
 * Copyright (C) 2002 Kelua SA
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Contact: jonathan@objectweb.org
 *
 * Author: Bruno Dumant
 *
 * -------------------------------------------------------------------
 * $ID$
 * -------------------------------------------------------------------
 * 
 */

package org.objectweb.jonathan.libs.kernel.decoders;

import org.objectweb.jonathan.apis.kernel.Context;
import org.objectweb.jonathan.apis.kernel.Component;
import org.objectweb.jonathan.apis.kernel.Name;
import org.objectweb.jonathan.apis.kernel.Factory;
import org.objectweb.jonathan.apis.kernel.MultiFactory;
import org.objectweb.jonathan.apis.kernel.Initializer;
import org.objectweb.jonathan.apis.kernel.Instanciable;
import org.objectweb.jonathan.apis.kernel.JonathanException;
import org.objectweb.jonathan.apis.kernel.ClassPathFinder;

import org.objectweb.jonathan.libs.helpers.HTable;
import org.objectweb.jonathan.libs.helpers.NameHelpers;
import org.objectweb.jonathan.libs.helpers.TraceHelpers;

import org.objectweb.jonathan.libs.kernel.JComponent;
import org.objectweb.jonathan.libs.kernel.ClassPathFinderImpl;

import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;
import java.util.List;
import java.util.ListIterator;


/**
 * Runtime implicit factory component. Provides the basic behaviour of a
 * <code>MultiFactory</code>: holds a set of implementation alternatives
 * (given at the object's creation) and instantiates one of them, according
 * to a given chosen alternative ID.
 */
public class JImplicitFactory extends JComponent implements Factory, MultiFactory, Initializer {

   static Name CLASSPATH_FINDER_NAME = NameHelpers.newName("/classpath_finder",'/');
   
   HTable alternatives = null;
   int nb_alternatives;
   JAlternative single;
   List alternatives_list;

   /**
    * Creates a new implicit factory whose implementation alternatives
    * are listed in <B>_alternatives</B> as JAlternatives.
    */
   public JImplicitFactory(List _alternatives) {
      alternatives_list = _alternatives;
      nb_alternatives = _alternatives.size();
      if (nb_alternatives > 1) {
         alternatives = HTable.newTable();
         JAlternative $alternative = null;
         nb_alternatives = 0;
         for (ListIterator $iterator = _alternatives.listIterator(); $iterator.hasNext();) {
            $alternative = (JAlternative) $iterator.next();
            alternatives.put($alternative.id,$alternative);
            nb_alternatives++;
         }
      } else if (nb_alternatives == 1) {
         single = (JAlternative) _alternatives.get(0);
      } else {
         System.err.println("WARNING in implicit factory [" +
                            reference +
                            "] : no alternative defined.");
      }
      value = this;
      type = JImplicitFactory.class;
   }

   /**
    * Creates a new implicit factory holding the same implementation
    * alternatives.
    */
   public Component duplicate() {
      return new JImplicitFactory(alternatives_list);
   }
   
   /**
    * From Factory interface.
    * Returns a new instance of the single alternative's implementation
    * by calling its constructor; <code>_c</code> is the local Context
    * containing the required instanciation parameters.
    * @param _c contains the instanciation parameter
    * @return an new instance of the implicit factory's single alternative
    * @exception JonathanException if something goes wrong.
    */
   public Object newObject(Context _c) throws JonathanException {
      return newObject(_c,null);
   }

   /**
    * Returns a new instance of the implicit factory's implementation
    * alternative whose ID is <code>_alternative</code>. In the single implementation
    * alternative case, it works only if the single alternative's ID is the
    * given one (although <code>newObject(Context)</code> would do it anyway).
    * @param _c where the instanciation parameters are found
    * @param _alternative the alternative to instantiate
    * @param _instance the method's caller.
    * @return an new <code>_alternative</code>'s instance
    * @exception JonathanException if something goes wrong.
    */
    public Object newObject(Context _c,int _alternative,Instanciable _instance) throws JonathanException {
      if (nb_alternatives == 1 && _alternative == single.id) {
          return newObject(_c,_instance);

      } else {
         JAlternative $alternative;
         if (alternatives == null ||
               ($alternative = (JAlternative) alternatives.get(_alternative)) == null) {
            throw new JonathanException("ERROR in implicit factory [" +
                                        reference +
                                        "] : alternative " + _alternative +
                                        " not found.");
         }
         
         Constructor $constructor = $alternative.constructor;
         if ($constructor == null) {
            Component context = reference.getContainer();
            $constructor = DecoderHelper.getConstructor($alternative.implementation_class,$alternative.arguments,context);
            $alternative.constructor = $constructor;
         }
         return DecoderHelper.apply($constructor,$alternative.arguments,_c,_instance);
      }
   }
   
   /**
    * Implementation of the Initializer interface's initialize method. 
    * Initializes an instance of the implicit factory's implementation
    * alternative whose ID is <code>_alternative</code> by calling its
    * setters. In the single implementation
    * alternative case, it works only if the single alternative's ID is the
    * given one (although <code>initialize(Object,Context)</code> would do it
    * anyway).
    * @param _obj the object to initialize
    * @param _c where the initialisation parameters are found
    * @param _alternative the alternative to instantiate
    * @exception JonathanException if something goes wrong.
    */
   public void initialize(Object _obj,Context _c,int _alternative) throws JonathanException {
      if (_obj == null) return;
      if (nb_alternatives == 1 && _alternative == single.id) {
         initialize(_obj,_c);
         return;
      } else if (_alternative == -1){
         initialize(_obj,_c);
         return;
      } else {
         JAlternative $alternative = (JAlternative) alternatives.get(_alternative);
         if ($alternative == null) {
            throw new JonathanException("ERROR in implicit factory [" +
                                        reference +
                                        "] : alternative " + _alternative +
                                        " not found.");
         }
         Component context = reference.getContainer();
         List $meths = $alternative.setters;
         for (int i=0; i<$meths.size(); i++){
            JSetter $current = (JSetter)$meths.get(i);
            Method $meth = DecoderHelper.getMethod(_obj.getClass(),$current.name,$current.args,context);
            Object $returnedByMeth = DecoderHelper.apply(_obj,$meth,$current.args,_c,null);
         }
      }
   }
      
   public String toString() {
      return "JImplicitFactory[" + alternatives + "]";
   }

   /**
    * Returns a new instance of the single alternative's implementation
    * by calling its constructor; <code>_c</code> is the local Context
    * containing the required instanciation parameters.
    * @param _c contains the instanciation parameter
    * @param _instance the method's caller.
    * @return an new instance of the implicit factory's single alternative
    * @exception JonathanException if something goes wrong.
    */
   private Object newObject(Context _c,Instanciable _instance) throws JonathanException {
      if (nb_alternatives == 1) {
         Constructor $constructor = single.constructor;
         if ($constructor == null) {
            Component context = reference.getContainer();
            $constructor = DecoderHelper.getConstructor(single.implementation_class,single.arguments,context);
            single.constructor = $constructor;
         }
         return DecoderHelper.apply($constructor,single.arguments,_c,_instance);
      } else {
         throw new JonathanException("ERROR in implicit factory [" +
                                     reference +
                                     "] : alternative choice required.");
      }   
   }

   /**
    * Works only in the single implementation alternative case.
    * Initialize an instance of the single alternative's implementation
    * by calling its setters; <code>_c</code> is the local Context
    * containing the required initialization parameters.
    * @param _obj an instance of the implicit factory's single alternative
    * @param _c contains the  initialization parameter
    * @exception JonathanException if something goes wrong.
    */
   private void initialize(Object _obj,Context _c) throws JonathanException {
      if (_obj == null) return;
      if (nb_alternatives == 1) {
         List $meths = single.setters;
         if (($meths == null) || ($meths.size() < 1)){
            return;
         }
         Component context = reference.getContainer();
         for (int i=0; i<$meths.size(); i++){
            JSetter $current = (JSetter)$meths.get(i);
            Method $meth = DecoderHelper.getMethod(_obj.getClass(),$current.name,$current.args,context);
            Object $returnedByMeth = DecoderHelper.apply(_obj,$meth,$current.args,_c,null);
         }
      } else {
         throw new JonathanException("ERROR in implicit factory [" +
                                     reference +
                                     "] : alternative choice required.");
      }
   }

}
