/***
 * Jonathan: an Open Distributed Processing Environment 
 * Copyright (C) 2001 Kelua SA
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Contact: jonathan@objectweb.org
 *
 * Author: Bruno Dumant
 * 
 */

package org.objectweb.jonathan.tools.compilers;

import org.objectweb.jonathan.tools.Kcf2java;
import nanoxml.XMLElement;
import nanoxml.XMLParseException;

import java.io.IOException;
import java.io.Writer;

/**
 * XML Property tag to runtime component java code compiler.
 * Used by {@link Kcf2java} through its <B>getTagCompiler()</B>
 * method to write Property tags' java code compilation.
 */
public class PropertyCompiler implements TagCompiler {
   
   static int nb = 0;

   /**
    * Translates a XML <code>PROPERTY</code> element to the java code
    * creating its corresponding property runtime component.
    * Supported property types are:
    * <UL><LI>String</LI>
    * <LI>char</LI>
    * <LI>boolean</LI>
    * <LI>byte</LI>
    * <LI>short</LI>
    * <LI>int</LI>
    * <LI>long</LI>
    * <LI>float</LI>
    * <LI>double</LI></UL>
    * @param _current_component the Property XML tag to be compiled
    * @param _element_name its element name
    * @param _configuration_var_name its containing configuration
    * java code variable name
    * @param _writer where to write the translated java code
    * @param _configuration_name the property's configuration
    * name XML property
    */
   public String writeComponent(XMLElement _current_component,
                                String _element_name,
                                String _configuration_var_name,
                                Writer _writer,
                                String _configuration_name)
      throws IOException {
      String $current_tag = _current_component.getName();
      
      if ($current_tag.equals("PROPERTY")){
         String $var_name = "PROPERTY" + (nb++);

         String $element_type = _current_component.getStringAttribute("TYPE");
         String $element_value = _current_component.getStringAttribute("VALUE");
         
         if ($element_type.equals("String")){
            _writer.write("         Component " + $var_name +
                          " = new JComponent(String.class,\"" + 
                          Kcf2java.normalize($element_value)+ "\");\n");

         } else if ($element_type.equals("char") ||
                    $element_type.equals("Character")){
            if ($element_value == null || $element_value.length() != 1) {
                throw new IOException("Value " + $element_value + " of element " + 
                                     _element_name + 
                                     " on line " + _current_component.getLineNr() + 
                                     " must be parsable as a character.");
            }
            _writer.write("         Component " + $var_name +
                          " = new JComponent(Character.class,new Character('" + 
                          Kcf2java.normalize($element_value)+ "'));\n");

         } else if ($element_type.equals("boolean") ||
                    $element_type.equals("Boolean")){
            if ($element_value.equalsIgnoreCase("true")) {
               _writer.write("         Component " + $var_name +
                             " = new JComponent(Boolean.class,Boolean.TRUE);\n");
            } else if ($element_value.equalsIgnoreCase("false")) {
               _writer.write("         Component " + $var_name +
                             " = new JComponent(Boolean.class,Boolean.FALSE);\n");
            } else {
               throw new IOException("Value " + $element_value + " of element " + 
                                     _element_name +
                                     " on line " + _current_component.getLineNr() + 
                                     " must be either 'true' or 'false'.");
            }

         } else if ($element_type.equals("byte") ||
                    $element_type.equals("Byte")){
            try {
               Byte $b = new Byte($element_value);
            } catch (NumberFormatException e) {
               throw new IOException("Value " + $element_value + " of element " + 
                                     _element_name + 
                                     " on line " + _current_component.getLineNr() + 
                                     " must be parsable as a byte.");
            }
            _writer.write("         Component " + $var_name +
                          " = new JComponent(Byte.class,new Byte(" + 
                          $element_value + "));\n");

         } else if ($element_type.equals("short") ||
                    $element_type.equals("Short")){
            try {
               Short $s = new Short($element_value);
            } catch (NumberFormatException e) {
               throw new IOException("Value " + $element_value + " of element " + 
                                     _element_name + 
                                     " on line " + _current_component.getLineNr() + 
                                     " must be parsable as a short.");
            }
            _writer.write("         Component " + $var_name +
                          " = new JComponent(Short.class,new Short(" + 
                          $element_value + "));\n");

         } else if ($element_type.equals("int") ||
                    $element_type.equals("Integer")){
            try {
               Integer $i = new Integer($element_value);
            } catch (NumberFormatException e) {
               throw new IOException("Value " + $element_value + " of element " + 
                                     _element_name + 
                                     " on line " + _current_component.getLineNr() + 
                                     " must be parsable as an integer.");
            }
            _writer.write("         Component "+ $var_name +
                          " = new JComponent(int.class," + $element_value + 
                          ");\n");

         } else if ($element_type.equals("long") ||
                    $element_type.equals("Long")){
            try {
               Long $l = new Long($element_value);
            } catch (NumberFormatException e) {
               throw new IOException("Value " + $element_value + " of element " + 
                                     _element_name + 
                                     " on line " + _current_component.getLineNr() + 
                                     " must be parsable as a long.");
            }
            _writer.write("         Component " + $var_name +
                          " = new JComponent(Long.class,new Long(" + 
                          $element_value + "));\n");

         } else if ($element_type.equals("float") ||
                    $element_type.equals("Float")){
            try {
               Float $f = new Float($element_value);
            } catch (NumberFormatException e) {
               throw new IOException("Value " + $element_value + " of element " + 
                                     _element_name + 
                                     " on line " + _current_component.getLineNr() + 
                                     " must be parsable as a float.");
            }
            _writer.write("         Component " + $var_name +
                          " = new JComponent(Float.class,new Float(" + 
                          $element_value + "));\n");

         } else if ($element_type.equals("double") ||
                    $element_type.equals("Double")){
            try {
               Double $d = new Double($element_value);
            } catch (NumberFormatException e) {
               throw new IOException("Value " + $element_value + " of element " + 
                                     _element_name + 
                                     " on line " + _current_component.getLineNr() + 
                                     " must be parsable as a double.");
            }
            _writer.write("         Component " + $var_name +
                          " = new JComponent(Double.class,new Double(" + 
                          $element_value + "));\n");

         } else if ($element_type.equals("null")){
            _writer.write("         Component " + $var_name +
                          " = new JComponent(Object.class,null);\n");

         } else {
            throw new IOException ("For the time being, only types int, Float, Boolean and String are \n" +
                                   "supported in a configuration file, see at line "+
                                   _current_component.getLineNr()+ " where type "+$element_type+" is used.");
         }
         return $var_name;
      } else {
         throw new IOException("TAG error at line " + _current_component.getLineNr() + ":\n" +
                               "PROPERTY excepted instead of " + $current_tag);
      }      
   }
}
